from abc import ABC, abstractmethod

from flloat.base.Alphabet import Alphabet
from flloat.syntax.ldlf import LDLfFormula

from pythomata.base.Symbol import Symbol
from typing import Set

from rltg.agents.feature_extraction import FeatureExtractor
from rltg.logic.RewardAutomaton import RewardAutomaton
from rltg.logic.RewardAutomatonSimulator import RewardAutomatonSimulator


class TemporalEvaluator(ABC):
    def __init__(self, goal_feature_extractor:FeatureExtractor, alphabet:Set[Symbol], formula:LDLfFormula, reward,
                 on_the_fly=False):
        self.goal_feature_extractor = goal_feature_extractor
        self.alphabet = Alphabet(alphabet)
        self.formula = formula
        if not on_the_fly:
            self._automaton = RewardAutomaton._fromFormula(alphabet, formula, reward)
            self.simulator = RewardAutomatonSimulator(self._automaton)
        else:
            self.simulator = self.formula.to_automaton(alphabet, on_the_fly=True)

    @abstractmethod
    def fromFeaturesToPropositional(self, features) -> Set[Symbol]:
        raise NotImplementedError

    def update(self, state):
        """update the automaton.
        :returns (new_automaton_state, reward)"""
        features = self.goal_feature_extractor(state)
        propositional = self.fromFeaturesToPropositional(features)
        reward = self.simulator.make_transition(propositional)
        return self.simulator.cur_state, reward

    def get_state(self):
        return self.simulator.cur_state

    def reset(self):
        self.simulator.reset()



import random

from gym import Space

from rltg.agents.exploration_policies.ExplorationPolicy import ExplorationPolicy


class RandomPolicy(ExplorationPolicy):

    def __init__(self, action_space:Space, epsilon_start=1.0, epsilon_end=0.1, decaying_steps=20000):
        super().__init__()
        self.action_space = action_space
        self.epsilon_start = epsilon_start
        self.epsilon_end = epsilon_end
        self.exploration_steps = decaying_steps

        self.epsilon = self.epsilon_start
        self.epsilon_decay_step = (self.epsilon_start - self.epsilon_end) \
                                  / self.exploration_steps

    def update(self, *args):
        if self.epsilon > self.epsilon_end:
            self.epsilon -= self.epsilon_decay_step
        else:
            self.epsilon = self.epsilon_end

    def explore(self, *args):
        action_id = None
        if random.random() < self.epsilon:
            action_id =self.action_space.sample()
        return action_id

    def reset(self, *args):
        pass



from flloat.base.Alphabet import Alphabet
from flloat.base.Symbol import Symbol
from flloat.syntax.ldlf import LDLfFormula
from pythomata.base.DFA import DFA
from typing import Set


class RewardAutomaton(DFA):
    def __init__(self, dfa:DFA, alphabet:Alphabet, f:LDLfFormula, reward):
        super().__init__(
            dfa.alphabet,
            dfa.states,
            dfa.initial_state,
            dfa.accepting_states,
            dfa.transition_function
        )
        self._dfa = dfa
        self.alphabet = alphabet
        self.f = f
        self.reward = reward

    @staticmethod
    def _fromFormula(alphabet:Set[Symbol], f:LDLfFormula, reward):
        dfa = f.to_automaton(alphabet, determinize=True, minimize=True)
        return RewardAutomaton(dfa, dfa.alphabet, f, reward)

    def get_reward(self):
        return self.reward

    def complete(self):
        return RewardAutomaton(self._dfa.complete(), self.alphabet, self.f, self.reward)

    def minimize(self):
        return RewardAutomaton(self._dfa.minimize(), self.alphabet, self.f, self.reward)

    def trim(self):
        return RewardAutomaton(self._dfa.trim(), self.alphabet, self.f, self.reward)



from typing import Set

from flloat.semantics.pl import PLInterpretation
from pythomata.base.Simulator import DFASimulator
from pythomata.base.Symbol import Symbol
from pythomata.base.utils import Sink

from rltg.logic.RewardAutomaton import RewardAutomaton


class RewardAutomatonSimulator(DFASimulator):
    def __init__(self, dfa:RewardAutomaton):
        super().__init__(dfa)
        self.visited_states = {self.cur_state}

    def reset(self):
        super().reset()
        self.visited_states = {self.cur_state}

    def make_transition(self, s:Set[Symbol]):
        i = PLInterpretation(s)
        super().make_transition(i)
        reward = 0

        # get the reward only if it is the first time we visit the state
        if self.cur_state not in self.visited_states:
            # get the total reward if we are in the final state
            if self.is_true():
                reward = self.dfa.get_reward()
            else:
                # get a fraction of the total reward to facilitate exploration
                # if we are in the Sink state, fail (highly negative reward).
                # TODO: backpropagation of the reward
                if isinstance(self.id2state[self.cur_state], Sink):
                    reward = -10000
                else:
                    reward = self.dfa.get_reward()/10

            self.visited_states.add(self.cur_state)

        return reward





import shutil

import numpy as np
from gym import Env

from rltg.agents.RLAgent import RLAgent
from rltg.agents.TGAgent import TGAgent
from rltg.utils.Renderer import PixelRenderer
from rltg.utils.StatsManager import StatsManager

import os

def goal_perc_threshold(*args, **kwargs)->bool:
    goal_history = kwargs["goal_history"]
    if len(goal_history)<100:
        return False
    goal_percentage = np.mean(goal_history[-100:])*100
    return goal_percentage >= 97.00

def check_automata_in_final_state(*args, **kwargs)->bool:
    temporal_evaluators = kwargs["temporal_evaluators"]
    # all the automata has to be in their final state
    return all(t.simulator.is_true() for t in temporal_evaluators)



ID2ACTION = {0: 2, 1: 3}
class Trainer(object):
    def __init__(self, env:Env, agent:RLAgent, n_episodes=1000,
                 eval=False,
                 resume=False,
                 renderer:PixelRenderer=None,
                 stopping_conditions=(goal_perc_threshold, check_automata_in_final_state),
                 agent_data_dir="agent_data"):
        self.env = env
        self.agent = agent
        self.stopping_conditions = stopping_conditions
        self.n_episodes = n_episodes
        self.resume = resume
        self.eval = eval
        self.renderer = renderer

        if not self.resume:
            shutil.rmtree(agent_data_dir, ignore_errors=True)
            os.mkdir(agent_data_dir)

        self.agent_data_dir = agent_data_dir

        if self.eval:
            self.agent.set_eval(self.eval)

    def main(self):
        env = self.env
        agent = self.agent
        num_episodes = self.n_episodes
        steps = 0

        stats = StatsManager()

        if self.resume:
            agent.load(self.agent_data_dir)

        # Main training loop
        for ep in range(num_episodes):

            total_reward = 0
            steps = 0

            state = env.reset()
            done = False

            while not done:
                action = agent.act(state)
                state2, reward, done, info = env.step(action)
                agent.observe(state, action, reward, state2)
                agent.replay()

                # add the observed reward (including the automaton reward)
                try:
                    total_reward += agent.brain.obs_history[-1][2]
                except:
                    total_reward += reward


                state = state2

                steps += 1
                agent.update()

                if self.renderer:
                    self.renderer.update(env)

            stats.update(len(agent.brain.Q), total_reward, info["goal"])
            stats.print_summary(ep, steps, len(agent.brain.Q), total_reward, agent.exploration_policy.epsilon, info["goal"])

            # stopping conditions
            temporal_evaluators = agent.temporal_evaluators if isinstance(agent, TGAgent) else []
            if all([s(goal_history=stats.goals, temporal_evaluators=temporal_evaluators)
                    for s in self.stopping_conditions]):
                break

            agent.reset()
            if ep%100==0:
                agent.save(self.agent_data_dir)

        agent.save(self.agent_data_dir)
        stats.plot()
